import numpy as np
from pyfftw.interfaces import numpy_fft as fft

from unyt import UnitRegistry
import unyt as U
from unyt import unyt_array, unyt_quantity
from unyt.dimensions import length

from numba import jit, njit
import logging
from scipy.spatial import cKDTree as KDTree
from scipy.interpolate import interpn
import numexpr as ne

from .utils import FiniteDictionary, CriticalPoints, unravel_index, solve

# create a logging format
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# Create a stream handler
handler = logging.StreamHandler()
handler.setFormatter(formatter)
handler.setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)

# Add handler to logger
logger.addHandler(handler)


@jit
def copy_xyz(xyz_rela):
    '''Keep only the points found within 1pixel from their parent cell'''
    xyz_rel = np.ascontiguousarray(xyz_rela.T)
    shape = xyz_rel.shape[1:]

    ndim = xyz_rel.shape[0]
    xyz_copy = xyz_rel.reshape(ndim, -1)

    N = xyz_copy.shape[1]

    count = 0
    for ii in range(N):
        ok = True
        for idim in range(ndim):
            ok &= -1 < xyz_copy[idim, ii] < 1

        if ok:
            count += 1

    xyz = np.zeros((ndim, count), dtype=xyz_rel.dtype)
    mask = np.zeros(N, dtype=np.int8)

    count = 0
    for ii in range(N):
        ok = True
        for idim in range(ndim):
            ok &= -1 < xyz_copy[idim, ii] < 1
        mask[ii] = ok
        if ok:
            ijk = list(unravel_index(ii, shape))
            for idim in range(ndim):
                xyz[idim, count] = ijk[idim]
            xyz[:, count] += xyz_copy[:, ii]
            count += 1

    return xyz.T, mask


@njit
def distance2(a, b, N):
    """Compute the distance between two vectors with wrapping of size N."""
    d2 = 0
    for i in range(a.shape[0]):
        aa = a[i]
        bb = b[i]
        dx = aa - bb
        if dx < -N/2:
            d2 += (dx + N)**2
        elif dx > N/2:
            d2 += (dx - N)**2
        else:
            d2 += (dx)**2
    return d2


@njit
def _cleanup_pairs_KDTree(xyz, xc, kind, pairs, N, shape, grad_norm=None):
    skip = np.zeros(len(xyz), dtype=np.uint8)
    for ip in range(len(pairs)):
        i, j = pairs[ip]
        if not skip[i] and not skip[j] and kind[i] == kind[j]:
            # Keep point closest to center of cell
            _xc = xc[i]
            if grad_norm is None:
                di = np.sum((xyz[i] - _xc)**2)
                dj = np.sum((xyz[j] - _xc)**2)
            else:
                di = grad_norm[i]
                dj = grad_norm[j]
            if di > dj:
                skip[i] = True
            else:
                skip[j] = True

    return skip


def cleanup_pairs_KDTree(xyz, kind, data_shape, dmin, grad):
    npoint, ndim = xyz.shape
    N = data_shape[0]
    logger.debug('Building KDTree')
    # TODO: support non square domains
    if not np.all(np.asarray(data_shape) == data_shape[0]):
        raise Exception('All axis should have the same dimension.')
    if len(xyz) == 0:
        return np.ones(0, dtype=bool)
    tree = KDTree(xyz, boxsize=data_shape[0], copy_data=True)
    pairs = tree.query_pairs(dmin, p=np.inf, output_type='ndarray')
    logger.debug('Removing close pairs')
    xc = np.round(xyz + 0.5) - 0.5
    skip = _cleanup_pairs_KDTree(xyz, xc, kind, pairs, N, data_shape,
                                 np.linalg.norm(grad, axis=1)).astype(bool)
    return ~skip


@jit(nopython=True)
def cleanup_pairs_N2(xyz, kind, data_shape):
    N = data_shape[0]
    npoint, ndim = xyz.shape

    for i in range(npoint):
        p1 = xyz[i, :]
        k1 = kind[i]
        ijk1 = unravel_index(i, data_shape)
        d1 = distance2(p1, ijk1, N)
        if np.all(p1 == -1):
            continue
        for j in range(i+1, npoint):
            p2 = xyz[j, :]
            if k1 != kind[j] or np.all(p2 == -1):
                continue

            if distance2(p1, p2, N) < 1:
                ijk2 = unravel_index(j, data_shape)
                d2 = distance2(p2, ijk2, N)

                # Keep 2
                if d1 > d2:
                    xyz[i, :] = -1
                else:
                    xyz[j, :] = -1


class ExtremaFinder(object):
    """A class to smooth an extract extrema from a n-dimensional field.

    Parameters
    ----------
    data : n-dimensional array
        The input field as a ndim array.
    boxlen : float
        Size of the box. All scales are then given in the same unit as boxlen.
    cache_len : int, optional
        Size of the cache containing the smoothed field. Defaults to 2.
    clean_pairs_method : str, optional
        The method used to discard close extrema. Defaults to 'KDTree'.
        See notes.
    nthreads : int
        Number of threads to use for FFT, etc.
    logleve : int
        Set to a low value for more verbosity

    Notes
    -----
    When discarding extrema, there are three methods available:
    * 'none': do not discard close extrema
    * 'direct': use an approach based on the relative distance between extrema.
                This scales as N^2 where N is the number of extrema.
    * 'KDTree': same as previous but using a KDTree, resulting in better perfs
                for large number of extrema.
    The default option gives the best result for large number of extrema.
    If the number of extrema is small, switch back to the 'direct' method.
    """

    ndim = None    # Number of dimensions
    boxlen = None  # Size of the box in physical units
    data_smooth = None    # Cache containing the real-space smoothed fields
    data_smooth_f = None  # Cache containing the Fourier-space smoothed fields
    extrema = None        # Dictionary containing the extrema
    data_raw = None    # Real-space raw data
    data_raw_f = None  # Fourier-space raw data

    kgrid = None  # The grid of Fourier kx, ky, ... modes
    k2 = None     # The grid of Fourier k^2

    clean_pairs_methods = ['KDTree', 'none', 'direct']
    clean_pairs_method = None

    dmin = 1  # Minimum distance between 2 points to merge them

    FFT_args = None

    def __init__(self, data, cache_len=2, clean_pairs_method='KDTree',
                 nthreads=1, loglevel=None, boxlen=1):

        self.registry = reg = UnitRegistry()
        reg.add("pixel", base_value=float((U.Mpc * boxlen / data.shape[0]).to('m')),
                dimensions=length)

        if loglevel:
            logger.setLevel(loglevel)

        self.ndim = data.ndim
        self.boxlen = boxlen
        ndim = self.ndim
        self.nthreads = nthreads
        self.clean_pairs_method = clean_pairs_method

        self.FFT_args = dict(threads=nthreads)

        self.data_raw = data
        self.data_shape = data.shape

        # Initialize caches
        self.data_smooth = FiniteDictionary(cache_len)
        self.data_smooth_f = FiniteDictionary(cache_len)
        self.extrema = {}

        # Perform first FFT + prepare k grid
        self.build_kgrid()
        self.fft_forward()

        # Initialize placeholders
        shape = list(self.data_raw.shape)
        shape_f = list(self.data_raw_f.shape)
        self.grad = np.zeros([ndim] + shape,
                             dtype=self.data_raw.dtype)
        self.hess = np.zeros([ndim*(ndim+1)//2] + shape,
                             dtype=self.data_raw.dtype)
        self.curvature = np.zeros(shape, dtype=self.data_raw.dtype)

        self.grad_f = np.zeros([ndim] + shape_f,
                               dtype=self.data_raw_f.dtype)
        self.hess_f = np.zeros([ndim*(ndim+1)//2] + shape_f,
                               dtype=self.data_raw_f.dtype)

    def array(self, value, units):
        return unyt_array(value, units, registry=self.registry)

    def quantity(self, value, units):
        return unyt_quantity(value, units, registry=self.registry)

    def build_kgrid(self):
        """Build the grid of k coefficients.

        Note
        ----
        This is made tricky by the FFT implementation. Usually, the
        last dimension is roughly half the input size (and complex),
        so the code has to take that into account.

        The last dimension is compressed because for real-input
        fields, all the information is stored in the complex part
        (twice the info of a real), so the shape is half the input
        size.
        """
        twopi = 2*np.pi
        shape = self.data_shape
        kall = ([np.fft.fftfreq(_)*twopi for _ in shape[:-1]] +
                [np.fft.rfftfreq(_)*twopi for _ in shape[-1:]])
        self.kgrid = np.asarray(np.meshgrid(*kall, indexing='ij'))
        self.k2 = (self.kgrid**2).sum(axis=0)

    def fft_forward(self):
        """Compute the direct Fourier transform of the input data."""
        logger.debug('Computing FFT of input with %s threads', self.nthreads)
        data = self.data_raw

        # Note: we don't use the defaults argument for the FFT as we
        # don't want any planning effort going to this one
        self.data_raw_f = fft.rfftn(data, threads=self.nthreads)

    def smooth(self, R):
        """Smooth the data at scale R (in pixel unit)."""
        if isinstance(R, unyt_quantity):
            R = float(R.to('pixel'))
        if R in self.data_smooth:
            return self.data_smooth[R]

        logger.debug('Smoothing at scale %.3f', R)
        data_f = self.data_raw_f * np.exp(-self.k2 * R**2 / 2)
        self.data_smooth_f[R] = data_f
        self.data_smooth[R] = fft.irfftn(data_f, **self.FFT_args)
        return self.data_smooth[R]

    def clean_pairs(self, xyz0, kind0, grad):
        if self.clean_pairs_method == 'KDTree':
            boxlen = self.data_raw.shape
            mask = cleanup_pairs_KDTree(
                np.mod(xyz0, boxlen),
                kind0, self.data_raw.shape, self.dmin, grad)

        elif self.clean_pairs_method == 'direct':
            cleanup_pairs_N2(xyz0, kind0, self.data_raw.shape)
            mask = np.any(xyz0 != -1, axis=1)

        elif self.clean_pairs_method == 'none':
            mask = np.ones(len(xyz0), dtype=bool)

        else:
            raise NotImplementedError(
                'The method %s is not implemented to clean '
                'pairs. Available methods are %s ' %
                (self.clean_pairs_method, self.clean_pairs_methods))

        return mask

    def compute_derivatives(self, R):
        """Compute the 1st and 2nd derivatives of the (Fourier) input.

        Argument
        --------
        R: float
            The smoothing scale.

        Return
        ------
        indices: (ndim, ndim, M)
            The M indices of the hessian. M = ndim*(ndim+1)/2.
        """
        if R not in self.data_smooth_f:
            self.smooth(R)

        data_f = self.data_smooth_f[R]
        ndim = self.ndim
        kgrid = self.kgrid
        ihess = 0
        indices = np.zeros((ndim, ndim), dtype=int)

        logger.debug('Computing hessian and gradient in Fourier space.')
        # Compute hessian and gradient in Fourier space
        for idim in range(ndim):
            k1 = kgrid[idim]
            self.grad_f[idim, ...] = ne.evaluate('data_f * 1j * k1')
            for idim2 in range(idim, ndim):
                k2 = kgrid[idim2]
                grad_f = self.grad_f[idim, ...]
                self.hess_f[ihess, ...] = ne.evaluate(
                    'grad_f * 1j * k2')

                indices[idim, idim2] = indices[idim2, idim] = ihess
                ihess += 1

        logger.debug('Inverse FFT of gradient')
        # Get them back in real space
        self.grad[...] = fft.irfftn(self.grad_f, axes=range(1, ndim+1), **self.FFT_args)
        logger.debug('Inverse FFT of hessian')
        self.hess[...] = fft.irfftn(self.hess_f, axes=range(1, ndim+1), **self.FFT_args)
        logger.debug('Curvature')
        if ndim == 3:
            a = self.hess[indices[0, 0]]
            b = self.hess[indices[1, 1]]
            c = self.hess[indices[2, 2]]
            d = self.hess[indices[0, 1]]
            e = self.hess[indices[0, 2]]
            f = self.hess[indices[1, 2]]
            self.curvature[...] = ne.evaluate(
                'a*b*c - c*d**2 - b*e**2 + 2*d*e*f - a*f**2',
                local_dict=dict(a=a, b=b, c=c, d=d, e=e, f=f))
        elif ndim == 2:
            a = self.hess[indices[0, 0]]
            b = self.hess[indices[1, 1]]
            c = self.hess[indices[0, 1]]
            self.curvature[...] = ne.evaluate(
                'a*b - c**2',
                local_dict=dict(a=a, b=b, c=c))
        else:
            self.curvature[...] = np.linalg.det(self.hess[indices, ...].T.copy()).T

        return indices

    def find_extrema(self, R):
        """Find the extrema of the field smoothed at scale R (in pixel unit).

        Return
        ------
        data: CriticalPoints
              The set of critical points found.
        """
        if isinstance(R, unyt_quantity):
            R = float(R.to('pixel'))
        if R in self.extrema:
            return self.extrema[R]
        ndim = self.ndim
        shape = self.data_shape

        if R not in self.data_smooth_f:
            self.smooth(R)

        indices = self.compute_derivatives(R)
        grid = np.linspace(0, shape[0], shape[0])

        rhs = -self.grad
        lhs = self.hess[indices.flatten(), ...].reshape(ndim, ndim, *shape)

        logger.debug('Solving linear system to find extrema')
        # Find the location of the points (in relative coordinates)
        # shape (N, N, N, ndim)
        xyz_rel = solve(np.asfortranarray(lhs.T),
                        np.asfortranarray(rhs.T))

        logger.debug('Discarding far extrema')
        # shapes (npoint, ndim) & (npoint, )
        xyz0, mask0 = copy_xyz(xyz_rel)
        mask0 = (mask0 == 1)

        # shape (npoint)
        dens0 = interpn([grid]*ndim, self.smooth(R), xyz0 % shape)

        # shape (npoint, ndim, ndim)
        # For the interpolation, move the hessian component at the end
        hess0 = (interpn([grid]*ndim, np.moveaxis(self.hess, 0, -1), xyz0 % shape)
                 [..., indices]
                 .reshape(-1, ndim, ndim))
        grad0 = (interpn([grid]*ndim, np.moveaxis(self.grad, 0, -1), xyz0 % shape)
                 .reshape(-1, ndim))

        logger.debug('Computing eigenvalues')
        # shape (npoint, ndim)
        eigvals0 = np.linalg.eigvalsh(hess0)

        # shape (npoint, )
        kind0 = (eigvals0 > 0).sum(axis=1)

        logger.info('Found %s extrema.', len(kind0))
        logger.debug('Cleaning up pairs with method %s',
                     self.clean_pairs_method)

        # Remove duplicate points
        mask = self.clean_pairs(xyz0, kind0, grad0)

        # Add units
        pos = self.array(xyz0[mask], 'pixel')
        hess = self.array(hess0[mask], '1/pixel**2')
        eigvals = self.array(eigvals0[mask], '1/pixel**2')
        dens = dens0[mask]
        sigma = self.smooth(R).std() * np.ones_like(dens)

        # Convert to physical units
        data = CriticalPoints(
            pos=pos,
            eigvals=eigvals,
            kind=kind0[mask],
            hessian=hess,
            npt=mask.sum(),
            dens=dens,
            sigma=sigma
        )

        self.extrema[R] = data
        return data



from numba import njit, jit, guvectorize
from collections import OrderedDict
import numpy as np
import numexpr as ne
import attr
import pandas as pd
from itertools import product
from scipy.interpolate import RegularGridInterpolator, interpn


class FiniteDictionary(OrderedDict):
    def __init__(self, maxlen, *args, **kwa):
        self._maxlen = maxlen
        super(FiniteDictionary, self).__init__(*args, **kwa)

    @property
    def maxlen(self):
        return self._maxlen

    @maxlen.setter
    def maxlen(self, v):
        if v < 0:
            raise Exception('Invalid maxlen %s', v)
        self._maxlen = v

    def __setitem__(self, k, v, *args, **kwa):
        if len(self) == self.maxlen and k not in self:
            # Remove oldest member
            self.popitem(False)

        super(FiniteDictionary, self).__setitem__(k, v, *args, **kwa)

    def last(self):
        return self[list(self.keys())[-1]]


def get_xyz_keys(Ndim):
    if Ndim <= 3:
        keys = ['x', 'y', 'z'][:Ndim]
    else:
        keys = [f'x{i+1}' for i in range(Ndim)]
    return keys


@attr.s(frozen=True)
class CriticalPoints:
    pos = attr.ib(converter=np.atleast_2d)
    eigvals = attr.ib(converter=np.atleast_2d)
    kind = attr.ib(converter=np.atleast_1d)
    hessian = attr.ib(converter=np.atleast_3d)
    npt = attr.ib(converter=int)
    dens = attr.ib(converter=np.atleast_1d)
    sigma = attr.ib(converter=np.atleast_1d)

    def as_dataframe(self):
        Ndim = self.pos.shape[1]
        keys = get_xyz_keys(Ndim)

        data = {}
        for i in range(Ndim):
            data[keys[i]] = self.pos[..., i]
            data[f'l{i+1}'] = self.eigvals[..., i]

            for j in range(i, Ndim):
                data[f'h{i+1}{j+1}'] = self.hessian[..., i, j]

        data['kind'] = self.kind
        data['dens'] = self.dens
        return pd.DataFrame(data)


@njit
def unravel_index(index, shape):
    n = len(shape)
    result = np.zeros(n, dtype=np.int32)
    for i in range(n-1, -1, -1):
        s = shape[i]
        result[i] = index % s
        index //= s

    return result


def solve(A, B):
    '''Solve the equation A*X = B.'''

    N = A.shape[-1]
    if N == 2:
        a = A[..., 0, 0]
        b = A[..., 1, 1]
        c = A[..., 0, 1]

        b1 = B[..., 0]
        b2 = B[..., 1]

        det = ne.evaluate('a*b - c**2')

        X = np.zeros(B.shape, order='F')
        X[..., 0] = ne.evaluate('(b*b1 - b2*c) / det')
        X[..., 1] = ne.evaluate('(a*b2 - b1*c) / det')

        return X

    elif N == 3:
        a = A[..., 0, 0]
        b = A[..., 1, 1]
        c = A[..., 2, 2]
        d = A[..., 0, 1]
        e = A[..., 0, 2]
        f = A[..., 1, 2]

        b1 = B[..., 0]
        b2 = B[..., 1]
        b3 = B[..., 2]

        d2 = d**2
        f2 = f**2
        e2 = e**2

        det = ne.evaluate('a*b*c - a*f2 - b*e2 - c*d2 + 2*d*e*f')

        X = np.zeros(B.shape, order='F')
        X[..., 0] = ne.evaluate('(b*b1*c - b2*c*d - b*b3*e + b3*d*f + b2*e*f - b1*f2) / det')
        X[..., 1] = ne.evaluate('(a*b2*c - b1*c*d + b3*d*e - b2*e2 - a*b3*f + b1*e*f) / det')
        X[..., 2] = ne.evaluate('(a*b*b3 - b3*d2 - b*b1*e + b2*d*e - a*b2*f + b1*d*f) / det')

        return X
    else:
        return np.linalg.solve(A, B)


@guvectorize(['void(float64[:], float64[:,:,:,:], float64[:])'],
             '(N),(M,i,i,i)->(M)')
def trilinear_interpolation(pos, v, ret):
    '''Compute the trilinear interpolation of data at given position

    Arguments
    ---------
    pos : (Ndim, ) float array
       The position w.r.t to the lower left edget of the cube.
    v : (M, 2, 2, 2) float array
       The value at the edges.

    Returns
    -------
    interp : float
       The interpolated value.

    Notes
    -----
    The original code is from
    http://paulbourke.net/miscellaneous/interpolation/

    '''
    xl, yl, zl = pos
    xr, yr, zr = 1-pos

    # Note the (inverse) order here!
    x = (xr, xl)
    y = (yr, yl)
    z = (zr, zl)
    ret[...] = 0

    for i in range(2):
        for j in range(2):
            for k in range(2):
                vol = x[i] * y[j] * z[k]
                ret[...] += v[..., i, j, k] * vol


@njit
def gradient(A, axis, dx=1):
    out = np.zeros_like(A)

    ijk = np.array([0, 0, 0], dtype=np.int32)
    ijkl = np.array([0, 0, 0], dtype=np.int32)
    ijkr = np.array([0, 0, 0], dtype=np.int32)

    i0 = j0 = k0 = 0
    iN, jN, kN = A.shape

    if axis == 0:
        i0 += 1
        iN -= 1
    elif axis == 1:
        j0 += 1
        jN -= 1
    elif axis == 2:
        k0 += 1
        kN -= 1

    for i in range(i0, iN):
        ijk[0] = ijkl[0] = ijkr[0] = i
        if axis == 0:
            ijkl[0] -= 1
            ijkr[0] += 1
        for j in range(j0, jN):
            ijk[1] = ijkl[1] = ijkr[1] = j
            if axis == 1:
                ijkl[1] -= 1
                ijkr[1] += 1
            for k in range(k0, kN):
                ijk[2] = ijkl[2] = ijkr[2] = k
                if axis == 2:
                    ijkl[2] -= 1
                    ijkr[2] += 1

                out[i, j, k] = (A[ijkr[0], ijkr[1], ijkr[2]] - A[ijkl[0], ijkl[1], ijkl[2]]) / 2 / dx

    # Left edge
    if axis == 0:
        i0 = 0
        iN = 1
    elif axis == 1:
        j0 = 0
        jN = 1
    elif axis == 2:
        k0 = 0
        kN = 1

    for i in range(i0, iN):
        ijk[0] = ijkr[0] = i
        if axis == 0:
            ijkr[0] += 1
        for j in range(j0, jN):
            ijk[1] = ijkr[1] = j
            if axis == 1:
                ijkr[1] += 1
            for k in range(k0, kN):
                ijk[2] = ijkr[2] = k
                if axis == 2:
                    ijkr[2] += 1

                out[i, j, k] = (A[ijkr[0], ijkr[1], ijkr[2]] - A[ijk[0], ijk[1], ijk[2]]) / dx

    # Right edge
    if axis == 0:
        i0 = A.shape[0]-1
        iN = A.shape[0]
    elif axis == 1:
        j0 = A.shape[1]-1
        jN = A.shape[1]
    elif axis == 2:
        k0 = A.shape[2]-1
        kN = A.shape[2]

    for i in range(i0, iN):
        ijk[0] = ijkl[0] = i
        if axis == 0:
            ijkl[0] -= 1
        for j in range(j0, jN):
            ijk[1] = ijkl[1] = j
            if axis == 1:
                ijkl[1] -= 1
            for k in range(k0, kN):
                ijk[2] = ijkl[2] = k
                if axis == 2:
                    ijkl[2] -= 1

                out[i, j, k] = (A[ijk[0], ijk[1], ijk[2]] - A[ijkl[0], ijkl[1], ijkl[2]]) / dx

    return out


@jit(looplift=True)
def measure_hessian_3d(position, data, LE=np.array([0, 0, 0])):
    '''Compute the value of the hessian of the field at the given position.

    Arguments
    ---------
    position : ndarray (Npt, Ndim)
       The position of the points in space in pixel units.
    data : ndarray (Npt, Npt, Npt)
       The field itself
    '''
    LE = np.asarray(LE)
    Npt = len(position)
    N = data.shape[0]
    Ndim = data.ndim

    buff = np.empty((6, 6, 6))
    # Contains the value of h_ij at the corner
    hij_buff = np.empty((6, 2, 2, 2))
    tmp_buff = np.empty((6, 6, 6))
    ret = np.empty((Npt, 3, 3))

    ipos = np.empty(Ndim, dtype=np.int32)
    jpos = np.empty(Ndim, dtype=np.int32)
    dpos = np.empty(Ndim, dtype=np.float64)

    for ipt in range(Npt):
        pos = position[ipt] - LE

        ipos[:] = pos-2
        jpos[:] = ipos+6
        dpos[:] = pos - ipos - 2

        # Copy data with periodic boundaries
        for i0, i in enumerate(range(ipos[0], jpos[0])):
            for j0, j in enumerate(range(ipos[1], jpos[1])):
                for k0, k in enumerate(range(ipos[2], jpos[2])):
                    buff[i0, j0, k0] = data[i % N, j % N, k % N]

        # Compute hessian using finite difference
        ii = 0
        for i in range(3):
            for jdim in range(i+1):
                tmp_buff[:] = gradient(gradient(buff, axis=i), axis=jdim)
                hij_buff[ii, :, :, :] = tmp_buff[2:4, 2:4, 2:4]

                ii += 1

        # Perform trilinear interpolation of the hessian
        tmp = trilinear_interpolation(dpos, hij_buff)
        ii = 0
        for i in range(3):
            for jdim in range(i+1):
                ret[ipt, i, jdim] = \
                  ret[ipt, jdim, i] = tmp[ii]
                ii += 1

    return ret


def measure_hessian(position, data, LE=np.array([0, 0, 0])):
    Ndim = data.ndim
    Npt = len(position)
    N = data.shape[0]

    if Ndim == 3:
        return measure_hessian_3d(position, data, LE)

    # Pad one in each dimension for the regular grid interpolation
    data_padded = np.pad(data, [(1, 1)]*Ndim, 'wrap')
    grid = [np.arange(-1, N+1)]*Ndim

    grad = [np.gradient(data_padded, axis=_) for _ in range(Ndim)]
    hess_flat = np.stack([
        np.gradient(grad[i], axis=j)
        for i in range(Ndim)
        for j in range(i, Ndim)], axis=-1)
    interpolator = RegularGridInterpolator(grid, hess_flat)

    # Unpack in Ndim x Ndim
    hess_interp = interpolator(position)
    hess_at_pt = np.empty((Npt, Ndim, Ndim))
    ii = 0
    for i in range(Ndim):
        for j in range(i, Ndim):
            hess_at_pt[:, i, j] = hess_at_pt[:, j, i] = hess_interp[:, ii]
            ii += 1
    return hess_at_pt


@jit(looplift=True)
def measure_gradient(position, data, LE=np.array([0, 0, 0])):
    '''Compute the value of the gradient of the field at the given position.

    Arguments
    ---------
    position : ndarray (Npt, Ndim)
       The position of the points in space in pixel units.
    data : ndarray (Npt, Npt, Npt)
       The field itself
    '''
    LE = np.asarray(LE)
    Npt = len(position)
    N = data.shape[0]

    buff = np.empty((4, 4, 4))
    # Contains the value of h_ij at the corner
    grad_buff = np.empty((3, 2, 2, 2))
    tmp_buff = np.empty((4, 4, 4))
    ret = np.empty((Npt, 3))

    ipos = np.empty(3, dtype=np.int32)
    jpos = np.empty(3, dtype=np.int32)
    dpos = np.empty(3, dtype=np.float64)

    for ipt in range(Npt):
        pos = position[ipt] - LE

        ipos[:] = pos-1
        jpos[:] = ipos+4
        dpos[:] = pos - ipos - 1

        # Copy data with periodic boundaries
        for i0, i in enumerate(range(ipos[0], jpos[0])):
            for j0, j in enumerate(range(ipos[1], jpos[1])):
                for k0, k in enumerate(range(ipos[2], jpos[2])):
                    buff[i0, j0, k0] = data[i % N, j % N, k % N]

        # Compute hessian using finite difference
        ii = 0
        for i in range(3):
            tmp_buff[:] = gradient(buff, axis=i)
            grad_buff[ii, :, :, :] = tmp_buff[1:3, 1:3, 1:3]

            ii += 1

        # Perform trilinear interpolation of the hessian
        tmp = trilinear_interpolation(dpos, grad_buff)
        ii = 0
        for i in range(3):
            ret[ipt, i] = tmp[ii]
            ii += 1

    return ret


def measure_third_derivative(position, data, eigenvectors):
    '''Measure the third derivative in the eigenframe

    Arguments
    ---------
    position : (N, Ndim)
    data : (M, M, M)
    eigenvectors : (N, Ndim, Ndim)

    Returns
    -------
    Fxii : (N, Ndim)
       The third derivative Fx11, Fx22, Fx33. See Notes for the
       definition of the "x" axis

    Notes
    -----
    The "x" axis is taken along the first eigenvector.
    '''
    N, Ndim = position.shape
    M = eigenvectors.shape[0]

    if eigenvectors.shape != (N, Ndim, Ndim):
        raise Exception(f'Wrong shape for eigenvectors: {eigenvectors.shape}, '
                        f'expected {(N, Ndim, Ndim)}.')
    if Ndim != 3:
        raise Exception(f'Got dimension {Ndim}, expected Ndim=3.')

    e1 = eigenvectors[..., 0]
    e2 = eigenvectors[..., 1]
    e3 = eigenvectors[..., 2]

    # Extend the field for interpolation
    F = data
    M = F.shape[0]
    Fext = np.zeros((M+2, M+2, M+2))
    x0 = position

    # Interpolate field along three eigenvectors
    a1, a2, a3 = np.meshgrid([-2, -1, 0, 1, 2], [-1, 0, 1], [-1, 0, 1],
                             indexing='ij')
    NA = np.newaxis
    xi = (a1[NA, :, :, :, NA] * e1[:, NA, NA, NA, :] +
          a2[NA, :, :, :, NA] * e2[:, NA, NA, NA, :] +
          a3[NA, :, :, :, NA] * e3[:, NA, NA, NA, :] +
          x0[:, NA, NA, NA, :]) % M

    # Extend field on boundaries for linear interpolation near edges
    sl1 = (slice(1, -1), slice(None))
    sl2 = (slice(1), slice(-1, None))
    sl3 = (slice(-1, None), slice(1))

    sl_all = (sl1, sl2, sl3)

    for (sx1, sx2), (sy1, sy2), (sz1, sz2) in product(*[sl_all]*3):
        Fext[tuple((sx1, sy1, sz1))] = F[tuple((sx2, sy2, sz2))]

    # Interpolate third derivative using second order finite
    # difference. Values has shape (3, Npt, 5)
    xx = np.arange(-1, M+1)
    v = interpn((xx, xx, xx), Fext, xi)

    D0 = [         0,  1,   0     ]
    D1 = [   0, -1/2,  0, 1/2,   0]  # 2nd order 1-derivative
    D2 = [         1, -2,   1     ]  # 2nd order 2-derivative
    D3 = [-1/2,    1,  0,  -1, 1/2]  # 2nd order 3-derivative

    # D1 = [ 1/12, -2/3, 0, 2/3, -1/12]  # 4th order 1-derivative
    F111 = np.einsum('...ijk,i,j,k', v, D3, D0, D0)
    F122 = np.einsum('...ijk,i,j,k', v, D1, D2, D0)
    F133 = np.einsum('...ijk,i,j,k', v, D1, D0, D2)

    return np.array((F111, F122, F133)).T



from scipy.spatial import cKDTree
from py_extrema.extrema import ExtremaFinder
import numpy as np
from tqdm.autonotebook import tqdm
import pandas as pd
from unyt import unyt_array

from .extrema import logger
from .utils import measure_hessian, measure_third_derivative, get_xyz_keys


class SloppingSaddle(object):
    """A class to detect slopping saddle point by successive smoothing.

    Parameters
    ----------
    extrema_finder : ExtremaFinder instance
    Rgrid : array like
        Smoothing scales, given in the same unit as the box size.
    """

    def __init__(self, extrema_finder, Rgrid):
        if not issubclass(type(extrema_finder), ExtremaFinder):
            raise Exception('Passed argument is not of type extrema finder.')

        self.ef = extrema_finder

        if not isinstance(Rgrid, unyt_array):
            Rgrid = self.ef.array(Rgrid, 'Mpc')
        self.Rgrid = Rgrid

    def compute_middle(self, x, y):
        boxlen = self.ef.data_shape[0]

        dx = x-y
        center = (x + y) / 2
        center[dx > +boxlen/2] -= boxlen / 2
        center[dx < -boxlen/2] += boxlen / 2

        return center % boxlen

    def detect_extrema(self, iRw=1):
        '''Compute the critical events found in a dataset.

        Parameters
        ----------
        iRw : int, default 1
            The thickness in smoothing scale dimension to look for critical
            points pair. See notes.
        ndim : int, optional, default 3

        Notes
        -----
        The algorithm first computes "heads" (critical points with no successor
        at next smoothing scale). It then tries to find pairs of heads of
        consecutive kind (e.g. peaks with filaments) between R[i] and R[i+iRw].
        '''
        ndim = self.ef.ndim

        # Extract relevant information from dsx
        dimensions = self.ef.data_raw.shape[0]
        smoothing_scales = self.Rgrid.to('pixel').value

        pos_keys = ['x', 'y', 'z'][:ndim]

        all_ext = []
        logger.debug('Building trees')
        for iR, R in enumerate(tqdm(smoothing_scales, desc='Building trees')):
            ext = self.ef.find_extrema(R).as_dataframe()
            ext['iR'] = iR
            ext['R'] = R
            all_ext.append(ext)

        # Build the dataframe containting the extrema
        ds = pd.concat(all_ext)
        ds['uid'] = np.arange(len(ds))
        ds['head'] = True
        ds = ds.reset_index().set_index(['iR', 'kind'])

        logger.debug('Locating heads')
        # Find all points that do not have a successor at a larger
        # smoothing scales ("heads")
        for kind in tqdm(range(ndim+1), leave=False):
            for iR in range(1, len(smoothing_scales)):
                if not ((iR, kind) in ds.index and (iR-1, kind) in ds.index):
                    continue
                p1 = ds.loc[(iR-1, kind), pos_keys].values % dimensions
                p2 = ds.loc[(iR,   kind), pos_keys].values % dimensions

                # Find elements of p2 in p1 (all should match!)
                t = cKDTree(p1, boxsize=dimensions)

                d, iprev = t.query(
                    p2,
                    distance_upper_bound=smoothing_scales[iR])
                ok = np.isfinite(d)
                head = np.ones(p1.shape[0], dtype=bool)
                head[iprev[ok]] = False

                ds.loc[(iR-1, kind), 'head'] = head

        # Build a tree out of the different heads
        heads = ds[ds['head']]

        pairs = []
        skip_uid = set()

        def kind_iter(iR, k0, k1, trees, slice_R):
            # Helper function that computes pairs of head
            h0 = heads.loc[(slice_R, k0), slice(None)]
            h1 = heads.loc[(slice_R, k1), slice(None)]
            if iRw == 1:
                p0 = h0[pos_keys].values
            else:
                p0 = h0[pos_keys + ['R']].values

            if len(h0) == 0 or len(h1) == 0:
                return [], [], []

            t1 = trees[k1]

            d, inext = t1.query(p0, distance_upper_bound=smoothing_scales[iR])
            ok = inext < len(t1.data)
            uid0 = h0[ok].uid.values
            uid1 = h1.iloc[inext[ok]]['uid'].values
            uids = np.sort(np.array((uid0, uid1)), axis=0)
            return uids[0], uids[1], d[ok] / smoothing_scales[iR]

        u0 = []
        u1 = []
        dist = []

        logger.debug('Locating critical events')
        for iR in tqdm(range(len(smoothing_scales)-iRw+1), leave=False):
            trees = {}
            slice_R = slice(iR, iR+iRw)

            # Look once in each direction (+kind, -kind)
            for kind in range(ndim+1):
                if iRw == 1:
                    p = heads.loc[(slice_R, kind), pos_keys].values
                    boxsize = [dimensions]*ndim
                else:
                    p = heads.loc[(slice_R, kind), pos_keys + ['R']].values
                    boxsize = [dimensions]*ndim + [smoothing_scales[-1]*100]
                if len(p) > 0:
                    # Wrap dimensions to prevent negative inputs
                    p = p % dimensions
                    trees[kind] = cKDTree(p, boxsize=boxsize)
            for kind in range(ndim):
                uid0, uid1, d = kind_iter(iR, kind, kind+1, trees, slice_R)
                u0.extend(uid0)
                u1.extend(uid1)
                dist.extend(d)

                uid0, uid1, d = kind_iter(iR, ndim-kind, ndim-kind-1, trees, slice_R)
                u0.extend(uid0)
                u1.extend(uid1)
                dist.extend(d)

        # Sort by increasing distances
        order = np.argsort(dist)
        u0 = np.array(u0)[order]
        u1 = np.array(u1)[order]
        dist = np.array(dist)[order]

        # Select each uid only once
        for uid0, uid1, d in zip(u0, u1, dist):
            if uid0 not in skip_uid or uid1 not in skip_uid and d < 1.5:
                pairs.append((uid0, uid1))
                skip_uid.add(uid0)
                skip_uid.add(uid1)

        # Given the merging pairs, compute the critical events
        uids = np.array(pairs).T
        ds_by_uid = ds.reset_index().set_index('uid')

        h0 = ds_by_uid.loc[uids[0]]
        h1 = ds_by_uid.loc[uids[1]]
        p1 = h0[pos_keys].values
        p2 = h1[pos_keys].values

        logger.debug('Computing critical point properties')
        data = {}
        for col in h0.columns:
            if h0[col].dtype == int:
                data[col] = (h0[col].values + h1[col].values) // 2
            else:
                data[col] = (h0[col].values + h1[col].values) / 2
        data['kind'] = (h0['kind'].values + h1['kind'].values - 1)//2
        L = dimensions
        # Compute position accounting for periodic boundaries
        pp = np.where(p1-p2 > L/2, (p1+p2-L)/2,
                      np.where(p1-p2 < -L/2,
                               (p1+p2+L)/2,
                               (p1+p2)/2))

        # Copy positional data
        for idim, pk in enumerate(pos_keys):
            data[pk] = pp[:, idim]

        critical_events = pd.DataFrame(data)

        self.critical_events = critical_events
        return self.critical_events
